# -*- coding: utf-8 -*-
# sourcery skip: avoid-builtin-shadow
"""
Created on 2022/09/18 17:04:28

@File -> mi_cmi.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 基于模型估计MI和CMI
"""

__doc__ = """
    使用模型近似估计互信息, 原理如下
    I(x;Y) = Accu(y, model.predict(x))
    I(x;Y|Z) = I(Y; xZ) - I(Y; Z)
             = Accu(y, model.predict(xz)) - Accu(y, model.predict(z))
"""

from sklearn.metrics import f1_score as cal_f1, r2_score as cal_r2
import numpy as np

from ...setting import DTYPES
from ...util import stdize_values


def _cal_metric(y_true, y_pred, ytype):
    if ytype == "d":
        return cal_f1(y_true, y_pred, average="micro")  # note 使用微平均
    else:
        return cal_r2(y_true, y_pred)


def _exec_model_test(x, y, ytype, model, test_ratio, z=None):
    N = x.shape[0]
    test_size = int(N * test_ratio)
    shuffled_indexes = np.random.permutation(range(N))
    train_idxs = shuffled_indexes[test_size:]
    test_idxs = shuffled_indexes[:test_size]

    x_train, x_test = x[train_idxs, :], x[test_idxs, :]
    y_train, y_test = y[train_idxs], y[test_idxs]

    model.fit(x_train, y_train)
    accu_x = _cal_metric(y_test, model.predict(x_test), ytype)

    if z is None:
        return accu_x
    
    xz = np.c_[x, z]
    xz_train, xz_test = xz[train_idxs, :], xz[test_idxs, :]
    model.fit(xz_train, y_train)
    accu_xz = _cal_metric(y_test, model.predict(xz_test), ytype)
    return accu_xz - accu_x


class MutualInfoModel(object):
    """基于模型的MI"""
    
    def __init__(self, x: np.ndarray, xtype: str, y: np.ndarray, ytype: str):
        assert xtype in DTYPES
        assert ytype in DTYPES
        self.x_norm = stdize_values(x, xtype)
        self.y_norm = stdize_values(y, ytype)
        self.xtype = xtype
        self.ytype = ytype
        
    def __call__(self, model, test_ratio=0.3):
        return _exec_model_test(
            self.x_norm, self.y_norm.flatten(), self.ytype, model, test_ratio)
        
        
class CondMutualInfoModel(object):
    """基于模型的CMI"""
    
    def __init__(self, x: np.ndarray, xtype: str, y: np.ndarray, ytype: str, z: np.ndarray, ztype: str):
        assert xtype in DTYPES
        assert ytype in DTYPES
        assert ztype in DTYPES
        self.x_norm = stdize_values(x, xtype)
        self.y_norm = stdize_values(y, ytype)
        self.z_norm = stdize_values(z, ztype)
        self.xtype = xtype
        self.ytype = ytype
        self.ztype = ztype
        
    def __call__(self, model, test_ratio=0.3):
        return _exec_model_test(
            self.x_norm, self.y_norm.flatten(), self.ytype, model, test_ratio, self.z_norm)
        

# if __name__ == "__main__":
#     def test_mi():
#         x = np.random.normal(0, 1, 10000)
#         y = np.random.randint(0, 5, 10000)
        
#         self = MutualInfoModel(x, "c", y, "d")
    
#         from lightgbm import LGBMClassifier as LightGBM
#         model = LightGBM(
#             n_estimators=100,
#             max_depth=3,
#             boosting_type="goss",
#             learning_rate=0.1,
#             importance_type="split",
#             n_jobs=3,
#         )
    
#         print(f"mi = {self(model)}")
    
#     def test_cmi():
#         x = np.random.normal(0, 1, 10000)
#         y = np.random.normal(0, 1, 10000)
#         z = np.random.normal(0, 1, 10000)
        
#         self = CondMutualInfoModel(x, "c", y, "c", z, "c")

#         from lightgbm import LGBMRegressor as LightGBM
#         model = LightGBM(
#             n_estimators=100,
#             max_depth=3,
#             boosting_type="goss",
#             learning_rate=0.1,
#             importance_type="split",
#             n_jobs=3,
#         )

#         print(f"cmi = {self(model)}")
        
#     test_mi()  # todo 分类问题算得关联值下限不为0
#     test_cmi()


# -*- coding: utf-8 -*-
# sourcery skip: avoid-builtin-shadow
"""
Created on 2021/08/07 16:38:23

@File -> mie_kraskov.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 使用Kraskov方法计算互信息
"""

from sklearn.neighbors import BallTree, KDTree
from scipy.special import digamma
import numpy.linalg as la
from numpy import log
import numpy as np
import warnings

__doc__ = """
    （补充文献）
"""


def entropy(x, k=3, base=2):
    """ The classic K-L k-nearest neighbor continuous entropy estimator
        x should be a list of vectors, e.g. x = [[1.3], [3.7], [5.1], [2.4]]
        if x is a one-dimensional scalar and we have four samples
    """
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    x = np.asarray(x)
    n_elements, n_features = x.shape
    x = add_noise(x)
    tree = build_tree(x)
    nn = query_neighbors(tree, x, k)
    const = digamma(n_elements) - digamma(k) + n_features * log(2)
    return (const + n_features * np.log(nn).mean()) / log(base)


def centropy(x, y, k=3, base=2):
    """ The classic K-L k-nearest neighbor continuous entropy estimator for the
        entropy of X conditioned on Y.
    """
    xy = np.c_[x, y]
    entropy_union_xy = entropy(xy, k=k, base=base)
    entropy_y = entropy(y, k=k, base=base)
    return entropy_union_xy - entropy_y


def tc(xs, k=3, base=2):
    xs_columns = np.expand_dims(xs, axis=0).T
    entropy_features = [entropy(col, k=k, base=base) for col in xs_columns]
    return np.sum(entropy_features) - entropy(xs, k, base)


def ctc(xs, y, k=3, base=2):
    xs_columns = np.expand_dims(xs, axis=0).T
    centropy_features = [centropy(col, y, k=k, base=base)
                         for col in xs_columns]
    return np.sum(centropy_features) - centropy(xs, y, k, base)


# def corex(xs, ys, k=3, base=2):
#     xs_columns = np.expand_dims(xs, axis=0).T
#     cmi_features = [mi(col, ys, k=k, base=base) for col in xs_columns]
#     return np.sum(cmi_features) - mi(xs, ys, k=k, base=base)


def kraskov_mi(x, y, z=None, k=3, base=np.e, alpha=0):
    """ Mutual information of x and y (conditioned on z if z is not None)
        x, y should be a list of vectors, e.g. x = [[1.3], [3.7], [5.1], [2.4]]
        if x is a one-dimensional scalar and we have four samples
    """
    assert len(x) == len(y), "Arrays should have same length"
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    x, y = np.asarray(x), np.asarray(y)
    x, y = x.reshape(x.shape[0], -1), y.reshape(y.shape[0], -1)
    x = add_noise(x)
    y = add_noise(y)
    points = [x, y]
    if z is not None:
        z = np.asarray(z)
        z = z.reshape(z.shape[0], -1)
        points.append(z)
    points = np.hstack(points)
    # Find nearest neighbors in joint space, p=inf means max-norm
    tree = build_tree(points)
    dvec = query_neighbors(tree, points, k)
    if z is None:
        a, b, c, d = avgdigamma(x, dvec), avgdigamma(
            y, dvec), digamma(k), digamma(len(x))
        if alpha > 0:
            d += lnc_correction(tree, points, k, alpha)
    else:
        xz = np.c_[x, z]
        yz = np.c_[y, z]
        a, b, c, d = avgdigamma(xz, dvec), avgdigamma(
            yz, dvec), avgdigamma(z, dvec), digamma(k)
    return (-a - b + c + d) / log(base)


# def cmi(x, y, z, k=3, base=2):
#     """ Mutual information of x and y, conditioned on z
#         Legacy function. Use mi(x, y, z) directly.
#     """
#     return mi(x, y, z=z, k=k, base=base)


def kldiv(x, xp, k=3, base=2):
    """ KL Divergence between p and q for x~p(x), xp~q(x)
        x, xp should be a list of vectors, e.g. x = [[1.3], [3.7], [5.1], [2.4]]
        if x is a one-dimensional scalar and we have four samples
    """
    assert k < min(len(x), len(xp)), "Set k smaller than num. samples - 1"
    assert len(x[0]) == len(xp[0]), "Two distributions must have same dim."
    x, xp = np.asarray(x), np.asarray(xp)
    x, xp = x.reshape(x.shape[0], -1), xp.reshape(xp.shape[0], -1)
    d = len(x[0])
    n = len(x)
    m = len(xp)
    const = log(m) - log(n - 1)
    tree = build_tree(x)
    treep = build_tree(xp)
    nn = query_neighbors(tree, x, k)
    nnp = query_neighbors(treep, x, k - 1)
    return (const + d * (np.log(nnp).mean() - np.log(nn).mean())) / log(base)


def lnc_correction(tree, points, k, alpha):
    e = 0
    n_sample = points.shape[0]
    for point in points:
        # Find k-nearest neighbors in joint space, p=inf means max norm
        knn = tree.query(point[None, :], k=k+1, return_distance=False)[0]
        knn_points = points[knn]
        
        # Substract mean of k-nearest neighbor points
        knn_points = knn_points - knn_points[0]
        
        # Calculate covariance matrix of k-nearest neighbor points, obtain eigen vectors
        covr = knn_points.T @ knn_points / k
        _, v = la.eig(covr)
        
        # Calculate PCA-bounding box using eigen vectors
        V_rect = np.log(np.abs(knn_points @ v).max(axis=0)).sum()
        
        # Calculate the volume of original box
        log_knn_dist = np.log(np.abs(knn_points).max(axis=0)).sum()

        # Perform local non-uniformity checking and update correction term
        if V_rect < log_knn_dist + np.log(alpha):
            e += (log_knn_dist - V_rect) / n_sample
    return e


# DISCRETE ESTIMATORS
def entropyd(sx, base=2):
    """ Discrete entropy estimator
        sx is a list of samples
    """
    unique, count = np.unique(sx, return_counts=True, axis=0)
    # Convert to float as otherwise integer division results in all 0 for proba.
    proba = count.astype(float) / len(sx)
    # Avoid 0 division; remove probabilities == 0.0 (removing them does not change the entropy estimate as 0 * log(1/0) = 0.
    proba = proba[proba > 0.0]
    return np.sum(proba * np.log(1. / proba)) / log(base)


def midd(x, y, base=2):
    """ Discrete mutual information estimator
        Given a list of samples which can be any hashable object
    """
    assert len(x) == len(y), "Arrays should have same length"
    return entropyd(x, base) - centropyd(x, y, base)


def cmidd(x, y, z, base=2):
    """ Discrete mutual information estimator
        Given a list of samples which can be any hashable object
    """
    assert len(x) == len(y) == len(z), "Arrays should have same length"
    xz = np.c_[x, z]
    yz = np.c_[y, z]
    xyz = np.c_[x, y, z]
    return entropyd(xz, base) + entropyd(yz, base) - entropyd(xyz, base) - entropyd(z, base)


def centropyd(x, y, base=2):
    """ The classic K-L k-nearest neighbor continuous entropy estimator for the
        entropy of X conditioned on Y.
    """
    xy = np.c_[x, y]
    return entropyd(xy, base) - entropyd(y, base)


def tcd(xs, base=2):
    xs_columns = np.expand_dims(xs, axis=0).T
    entropy_features = [entropyd(col, base=base) for col in xs_columns]
    return np.sum(entropy_features) - entropyd(xs, base)


def ctcd(xs, y, base=2):
    xs_columns = np.expand_dims(xs, axis=0).T
    centropy_features = [centropyd(col, y, base=base) for col in xs_columns]
    return np.sum(centropy_features) - centropyd(xs, y, base)


def corexd(xs, ys, base=2):
    xs_columns = np.expand_dims(xs, axis=0).T
    cmi_features = [midd(col, ys, base=base) for col in xs_columns]
    return np.sum(cmi_features) - midd(xs, ys, base)


# MIXED ESTIMATORS
def micd(x, y, k=3, base=2, warning=True):
    """ If x is continuous and y is discrete, compute mutual information
    """
    assert len(x) == len(y), "Arrays should have same length"
    entropy_x = entropy(x, k, base)

    y_unique, y_count = np.unique(y, return_counts=True, axis=0)
    y_proba = y_count / len(y)

    entropy_x_given_y = 0.
    for yval, py in zip(y_unique, y_proba):
        x_given_y = x[(y == yval).all(axis=1)]
        if k <= len(x_given_y) - 1:
            entropy_x_given_y += py * entropy(x_given_y, k, base)
        else:
            if warning:
                warnings.warn("Warning, after conditioning, on y={yval} insufficient data. "
                              "Assuming maximal entropy in this case.".format(yval=yval))
            entropy_x_given_y += py * entropy_x
    return abs(entropy_x - entropy_x_given_y)  # units already applied


def midc(x, y, k=3, base=2, warning=True):
    return micd(y, x, k, base, warning)


def centropycd(x, y, k=3, base=2, warning=True):
    return entropy(x, base) - micd(x, y, k, base, warning)


def centropydc(x, y, k=3, base=2, warning=True):
    return centropycd(y, x, k=k, base=base, warning=warning)


def ctcdc(xs, y, k=3, base=2, warning=True):
    xs_columns = np.expand_dims(xs, axis=0).T
    centropy_features = [centropydc(
        col, y, k=k, base=base, warning=warning) for col in xs_columns]
    return np.sum(centropy_features) - centropydc(xs, y, k, base, warning)


def ctccd(xs, y, k=3, base=2, warning=True):
    return ctcdc(y, xs, k=k, base=base, warning=warning)


def corexcd(xs, ys, k=3, base=2, warning=True):
    return corexdc(ys, xs, k=k, base=base, warning=warning)


def corexdc(xs, ys, k=3, base=2, warning=True):
    return tcd(xs, base) - ctcdc(xs, ys, k, base, warning)


# UTILITY FUNCTIONS

def add_noise(x, intens=1e-10):
    # small noise to break degeneracy, see doc.
    return x + intens * np.random.random_sample(x.shape)


def query_neighbors(tree, x, k):
    return tree.query(x, k=k + 1)[0][:, k]


def count_neighbors(tree, x, r):
    return tree.query_radius(x, r, count_only=True)


def avgdigamma(points, dvec):
    # This part finds number of neighbors in some radius in the marginal space
    # returns expectation value of <psi(nx)>
    tree = build_tree(points)
    dvec = dvec - 1e-15
    num_points = count_neighbors(tree, points, dvec)
    return np.mean(digamma(num_points))


def build_tree(points):
    if points.shape[1] >= 20:
        return BallTree(points, metric='chebyshev')
    return KDTree(points, metric='chebyshev')
    
    # if points.shape[1] >= 20:
    #     return BallTree(points, metric='minkowski')
    # return KDTree(points, metric='minkowski')

# TESTS


def shuffle_test(measure, x, y, z=False, ns=200, ci=0.95, **kwargs):
    """ Shuffle test
        Repeatedly shuffle the x-values and then estimate measure(x, y, [z]).
        Returns the mean and conf. interval ('ci=0.95' default) over 'ns' runs.
        'measure' could me mi, cmi, e.g. Keyword arguments can be passed.
        Mutual information and CMI should have a mean near zero.
    """
    x_clone = np.copy(x)  # A copy that we can shuffle
    outputs = []
    for _ in range(ns):
        np.random.shuffle(x_clone)
        if z:
            outputs.append(measure(x_clone, y, z, **kwargs))
        else:
            outputs.append(measure(x_clone, y, **kwargs))
    outputs.sort()
    return np.mean(outputs), (outputs[int((1. - ci) / 2 * ns)], outputs[int((1. + ci) / 2 * ns)])


# if __name__ == "__main__":
#     print("MI between two independent continuous random variables X and Y:")
#     np.random.seed(0)
#     x = np.random.randn(1000, 10)
#     y = np.random.randn(1000, 3)
#     print(mi(x, y, base=2, alpha=0))


# -*- coding: utf-8 -*-
"""
Created on 2022/09/18 15:52:28

@File -> mi.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 互信息
"""

from numpy import log
import numpy as np

from ....util import stdize_values
from ....setting import DTYPES, BASE
from ..entropy.marg_entropy import MargEntropy
from ._kraskov import kraskov_mi as cal_kraskov_mi
from ._ross import cal_ross_mi


def _cal_mi_cc(x, y, k):
    return cal_kraskov_mi(x, y, k=k, base=BASE, alpha=0)
    
    
def _cal_mi_cd(x, y, k, metric="chebyshev", base=BASE):
    return cal_ross_mi(x, y, k, metric, base)
    

def _cal_mi_dd(x, y):
    return MargEntropy(x, "d")() + MargEntropy(y, "d")() - MargEntropy(np.c_[x, y], "d")()


class MutualInfoGIEF(object):
    """两一维或多维变量间的互信息"""
    
    def __init__(self, x: np.ndarray, xtype: str, y: np.ndarray, ytype: str):
        assert xtype in DTYPES
        assert ytype in DTYPES
        self.x_norm = stdize_values(x, xtype)
        self.y_norm = stdize_values(y, ytype)
        self.xtype = xtype
        self.ytype = ytype
        
    def __call__(self, k: int=3, metric: str="chebyshev"):
        if (self.xtype == "d") & (self.ytype == "d"):
            return _cal_mi_dd(self.x_norm, self.y_norm) / log(BASE)
        elif (self.xtype == "c") & (self.ytype == "d"):
            return _cal_mi_cd(self.x_norm, self.y_norm, k, metric=metric) / log(BASE)
        elif (self.xtype == "d") & (self.ytype == "c"):
            return _cal_mi_cd(self.y_norm, self.x_norm, k, metric=metric) / log(BASE)
        elif (self.xtype == "c") & (self.ytype == "c"):
            return _cal_mi_cc(self.x_norm, self.y_norm, k) / log(BASE)
        else:
            raise RuntimeError("")


# -*- coding: utf-8 -*-
"""
Created on 2022/08/24 22:44:39

@File -> coefficient.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 相关系数计算
    note: 对类别变量进行编码, 然后视作数值变量计算相关系数
"""

from scipy.stats import pearsonr, spearmanr
import numpy as np
import dcor

from ...util import SuperCategorEncoding


def cal_dist_corr(x, y, x_type: str="c"):
    """距离相关系数"""
    if x_type == "d":
        x = _encode(x, y)
    return np.abs(dcor.distance_correlation(x, y))


def cal_pearson_corr(x, y, x_type: str="c"):
    """Pearson相关系数"""
    if x_type == "d":
        x = _encode(x, y)
    return np.abs(pearsonr(x, y)[0])


def cal_spearman_corr(x, y, x_type: str="c"):
    """Spearman相关系数"""
    if x_type == "d":
        x = _encode(x, y)
    return np.abs(spearmanr(x, y)[0])
    

def _encode(x, y):
    """如果x是类别型变量, 则对x进行编码
    注意: 这里选择有监督的编码,因此入参有y, 其他编码方式可以在univar_encoding里选择
    """
    super_enc = SuperCategorEncoding(x, y)
    return super_enc.mhg_encoding()


# -*- coding: utf-8 -*-
# sourcery skip: avoid-builtin-shadow
"""
Created on 2022/09/18 15:28:09

@File -> marg_entropy.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 边际熵
"""

__doc__ = """
    本代码用于对一维和多维离散或连续变量数据的信息熵和互信息进行计算.
    连续变量信息熵使用Kraskov等人的方法计算, Lord等人文献可作为入门;离散变量信息熵则直接进行计算.

    求得信息熵和互信息后, 便可对条件熵进行求取:

    H(Y|X) = H(Y) - I(X;Y)

    此处不再补充条件熵的相关代码.

    参考文献
    1. W. M. Lord, J. Sun, E. M. Bolt: Geometric K-Nearest Neighbor Estimation of Entropy and Mutual information, 2018.
    2. A. Kraskov, H. Stoegbauer, P. Grassberger: Estimating Mutual Information, 2003.
    3. D. Lombardi, S. Pant: A Non-Parametric K-Nearest Neighbor Entropy Estimator, 2015.
    4. B. C. Ross: Mutual Information between Discrete and Continuous Data Sets, 2014.
"""

from scipy.special import psi
from numpy import log
import numpy as np

from ....util import stdize_values, build_tree, query_neighbors_dist, get_unit_ball_volume
from ....setting import DTYPES, BASE


def _cal_discrete_entropy(x):
    _, counts = np.unique(x, return_counts=True, axis=0)
    proba = counts.astype(float) / len(x)
    proba = proba[proba > 0.0]
    return np.sum(proba * np.log(1. / proba))


def _cal_kl_entropy(x, k, metric="chebyshev"):
    assert k <= len(x) - 1
    N, D = x.shape
    
    # 构建距离树
    tree = build_tree(x, "chebyshev")

    # 计算结果
    nn_distc = query_neighbors_dist(tree, x, k)  # 获得了各样本第k近邻的距离
    v = get_unit_ball_volume(D, metric)
    return (-psi(k) + psi(N) + np.log(v) + D * np.log(nn_distc).mean())
    

class MargEntropy(object):
    """计算任意连续和离散变量的信息熵"""
    
    def __init__(self, x: np.ndarray, xtype: str):
        assert xtype in DTYPES
        self.x_norm = stdize_values(x, xtype)
        self.xtype = xtype
        
    def __call__(self, k: int=3, metric: str="chebyshev"):
        if self.xtype == "d":
            return _cal_discrete_entropy(self.x_norm) / log(BASE)
        elif self.xtype == "c":
            return _cal_kl_entropy(self.x_norm, k, metric) / log(BASE)


import numpy as np

from ...util import stdize_values
from ._quant_darbellay import exec_partition, Cell


class MutualInfoDarbellay(object):
    """基于Darbellay自适应分箱的互信息计算"""
    
    def __init__(self, x: np.ndarray, y: np.ndarray):
        # NOTE: 由于本项目中的Darbellay离散化算法设置, 需要对变量数据进行标准化至0-1区间
        self.x_norm = stdize_values(x, "c")
        self.y_norm = stdize_values(y, "c")
        try:
            assert self.x_norm.shape[1] == 1
            assert self.y_norm.shape[1] == 1
        except Exception as e:
            raise ValueError("不支持非一维变量间的MI计算") from e
    
    def __call__(self):
        leaf_cells, arr_norm = exec_partition(self.x_norm, self.y_norm)
        N_total, _ = arr_norm.shape
        
        # 计算互信息.
        n_leafs = len(leaf_cells)

        mi = 0.0
        for i in range(n_leafs):
            cell = leaf_cells[i]  # type: Cell
            (xl, xu), (yl, yu) = cell.bounds

            Nxy = len(cell.arr)
            Nx = len(
                np.where((arr_norm[:, 0] >= xl) & (arr_norm[:, 0] < xu))[0])
            Ny = len(
                np.where((arr_norm[:, 1] >= yl) & (arr_norm[:, 1] < yu))[0])
            gain = Nxy * np.log(Nxy / Nx / Ny)
            mi += gain
        
        mi = round(mi / N_total + np.log(N_total), 4)
        return mi
    

# if __name__ == "__main__":
    
#     def test_cc():
#         x = np.random.normal(0, 1, 10000)
#         y = np.random.normal(0, 1, 10000)
#         print(f"mi = {MutualInfoDarbellay(x, y)()}")
        
#     test_cc()


from typing import List
import numpy as np

from ...util import stdize_values

if "plt" not in dir():
    import matplotlib.pyplot as plt
    


def cal_area(bounds: List[tuple], D) -> float:
    """计算面积"""
    area = 1.0
    for i in range(D):
        bd = bounds[i]
        area *= max(bd) - min(bd)
    return area


def cal_equi_partition_thres(arr: np.ndarray) -> List[float]:
    """获取各维度上等边际概率(即等边际样本数)分箱的阈值, 维数为2
    :param arr: 待分析二维样本数组
    """
    # todo: 检查arr可否为空
    N, D = arr.shape
    
    # 如果只有1个样本, 就直接返回样本各维度的值
    if N == 1:
        return list(arr.flatten())
    
    # 如果多于1个样本, 则偶数时取中间两个值的均值, 奇数时取中间值两侧值的均值
    part_idx = N // 2  # 离散化位置idx
    part_thres = []
    for i in range(D):
        series_srt = np.sort(arr[:, i])  # 对应维度值升序排列
        if N % 2 == 0:  # 以均值划分
            marginal_part_value = (series_srt[part_idx - 1] + series_srt[part_idx]) / 2
        else:
            marginal_part_value = (series_srt[part_idx - 1] + series_srt[part_idx + 1]) / 2
        part_thres.append(marginal_part_value)
    return part_thres


class Cell(object):
    """样本数据格对象"""
    
    def __init__(self, arr: np.ndarray=None) -> None:
        # if (arr is None) or (len(arr) < N_min):
        if (arr is None) or (len(arr) == 0):
            self.is_empty = True
            self.arr = np.array([]).reshape(0, 2)
            self.N, self.D = self.arr.shape
        else:
            arr = arr.copy()
            self.is_empty = False
            self.arr = arr.reshape(len(arr), -1)
            self.N, self.D = self.arr.shape
            if self.D != 2:
                raise RuntimeError(f"现有代码仅支持两个一维数据间的计算, self.D=2, 但实际self.D={self.D}")
    
    def def_cell_bounds_cal_area(self, bounds: List[tuple]):
        """手动定义cell的边界并计算对应覆盖的面积
        :param bounds: 边界值list, 如[(x_min, x_max), (y_min, y_max)]
        """
        self.bounds = bounds
        self.area = cal_area(self.bounds, self.D)
    
    def cal_proba_dens(self, N_total: int) -> float:
        """计算以样本数计的概率密度"""
        return 0.0 if self.area == 0.0 else self.N / (N_total * self.area)

    def get_marginal_partition_thres(self):
        self.part_thres = cal_equi_partition_thres(self.arr)
        
    def exec_partition(self):
        """执行等频率离散化, 要求样本self.N > 0"""
        assert self.N > 0
        
        # 先在x方向上分为左右两部分.
        part_arr_l = self.arr[
            np.where((self.arr[:, 0] < self.part_thres[0]) &
                     (self.arr[:, 0] >= self.bounds[0][0]))
        ]
        part_arr_r = self.arr[
            np.where((self.arr[:, 0] >= self.part_thres[0])
                     & (self.arr[:, 0] <= self.bounds[0][1]))
        ]

        # 再在y方向上继续切分.
        part_arr_ul = part_arr_l[np.where(
            (part_arr_l[:, 1] >= self.part_thres[1]) & (part_arr_l[:, 1] <= self.bounds[1][1]))]
        part_arr_ll = part_arr_l[np.where(
            (part_arr_l[:, 1] < self.part_thres[1]) & (part_arr_l[:, 1] >= self.bounds[1][0]))]

        part_arr_ur = part_arr_r[np.where(
            (part_arr_r[:, 1] >= self.part_thres[1]) & (part_arr_r[:, 1] <= self.bounds[1][1]))]
        part_arr_lr = part_arr_r[np.where(
            (part_arr_r[:, 1] < self.part_thres[1]) & (part_arr_r[:, 1] >= self.bounds[1][0]))]

        cell_ul, cell_ur, cell_ll, cell_lr = Cell(part_arr_ul), Cell(part_arr_ur), \
            Cell(part_arr_ll), Cell(part_arr_lr)

        # 确定边界.
        (xl, xu), (yl, yu) = self.bounds
        x_thres, y_thres = self.part_thres
        cell_ul.def_cell_bounds_cal_area([(xl, x_thres), (y_thres, yu)])
        cell_ur.def_cell_bounds_cal_area([(x_thres, xu), (y_thres, yu)])
        cell_ll.def_cell_bounds_cal_area([(xl, x_thres), (yl, y_thres)])
        cell_lr.def_cell_bounds_cal_area([(x_thres, xu), (yl, y_thres)])
        return cell_ul, cell_ur, cell_ll, cell_lr
    
    def show(self, linewidth: float = 0.5):
        (xl, xu), (yl, yu) = self.bounds
        plt.plot([xl, xu], [yl, yl], "-", c="k", linewidth=linewidth)
        plt.plot([xu, xu], [yl, yu], "-", c="k", linewidth=linewidth)
        plt.plot([xu, xl], [yu, yu], "-", c="k", linewidth=linewidth)
        plt.plot([xl, xl], [yu, yl], "-", c="k", linewidth=linewidth)
        

# 递归离散化.
def _try_partition(cell: Cell, p_eps: float, N_total: int):
    """尝试对Cell进行一次分裂, 若样本数少于规定阈值或概率密度收敛, 则返回为四个空cell, 否则正常分裂"""
    # 尝试分裂一下, 并检查分裂效果.
    proba_dens = cell.cal_proba_dens(N_total)

    cell.get_marginal_partition_thres()
    cell_ul, cell_ur, cell_ll, cell_lr = cell.exec_partition()

    # 所有子cell的PDF均收敛则为收敛
    is_proba_dens_converged = all(np.abs(c.cal_proba_dens(N_total) - proba_dens) / proba_dens <= p_eps \
        for c in [cell_ul, cell_ur, cell_ll, cell_lr])

    return (Cell(), Cell(), Cell(), Cell()) if is_proba_dens_converged else (cell_ul, cell_ur, cell_ll, cell_lr)


def recursively_partition(cell: Cell, N_total: int, min_samples_split: int = 30, p_eps: float = 1e-3):
    """对一个cell进行递归离散化

    :param cell: 初始cell
    :param p_eps: 子cell概率与父cell相对偏差阈值, 如果所有都小于该值则终止离散化, defaults to 1e-3
    """
    leaf_cells = []

    def _partition(cell):
        if cell.N >= min_samples_split:
            part_ul, part_ur, part_ll, part_lr = _try_partition(cell, p_eps, N_total)
            
            if all(p.is_empty for p in [part_ul, part_ur, part_ll, part_lr]):
                leaf_cells.append(cell)
            else:
                _partition(part_ul)
                _partition(part_ur)
                _partition(part_ll)
                _partition(part_lr)
        else:
            leaf_cells.append(cell)

    _partition(cell)
    return leaf_cells


def exec_partition(x: np.ndarray, y: np.ndarray):
    """对X和Y数据在二维平面进行离散化"""
    x, y = x.flatten(), y.flatten()
    assert len(x) == len(y)
    N_total = len(x)
    
    # 2.1 预处理
    arr = np.c_[x, y]
    arr_norm = stdize_values(arr, dtype="c")
    
    # 2.2 初始化Cell
    cell = Cell(arr_norm)
    cell.def_cell_bounds_cal_area([(0.0, 1.0), (0.0, 1.0)])
    
    # 2.3 执行自适应划分
    leaf_cells = recursively_partition(cell, N_total, min_samples_split=30, p_eps=1e-3)
    
    return leaf_cells, arr_norm


from scipy.special import psi
from numpy import log
import numpy as np

from ....util import build_tree, query_neighbors_dist


# ---- 高维数据压缩, 计算互信息时要用. ----------------------------------------------------------------

def _reorder_z_series(z_compress: np.ndarray):
    z_compress = z_compress.copy()
    map_ = dict(zip(set(z_compress), list(range(len(set(z_compress))))))
    vfunc = np.vectorize(lambda x: map_[x])
    z_compress = vfunc(z_compress)
    return z_compress


def compress_z_data(z_arr: np.ndarray):
    """对高维数据进行压缩
    :param z_arr: Z数组, 注意shape = (D, N)而不是(N, D)
    """
    D = z_arr.shape[0]
    z_label_ns = [len(set(z_arr[i, :])) for i in range(z_arr.shape[0])]

    if D == 0:
        raise ValueError('empty array z')

    if D == 1:
        z_compress = z_arr.flatten()
        z_compress = _reorder_z_series(z_compress)
    else:
        i = 2
        z_compress = np.ravel_multi_index(z_arr[:i, :], z_label_ns[:i], mode='wrap')
        z_compress = _reorder_z_series(z_compress)

        while i != D:
            _arr = np.vstack((z_compress, z_arr[i, :]))
            z_compress = np.ravel_multi_index(_arr, (len(set(z_compress)), z_label_ns[i]), mode='wrap')
            z_compress = _reorder_z_series(z_compress)
            i += 1

    return z_compress


def cal_ross_mi(x, y, k, metric="chebyshev", base=np.e):
    x, y = x.copy(), y.copy()
    N = x.shape[0]
    
    # Note: 对高维y进行重编码, 压缩为1维序列
    y = y.reshape(y.shape[0], -1)
    if y.shape[1] > 1:
        y = compress_z_data(y.T).reshape(-1, 1)
        
    assert (x.shape[1] >= 1) & (y.shape[1] == 1)  # NOTE: 此处y必须为1维
    y = y.flatten()

    k = 3 if k is None else k
    assert k <= N - 1

    # 统计各类Y的总数.
    classes = np.unique(y)
    Nx_class = np.zeros_like(y)
    for i, _ in enumerate(y):
        Nx_class[i] = np.sum(y == y[i])

    # 逐类进行K近邻计算.
    nn_distc_classes = np.zeros_like(y, dtype=float)
    for c in classes:
        mask = np.where(y == c)[0]
        tree = build_tree(x[mask, :], "chebyshev")
        nn_distc_classes[mask] = query_neighbors_dist(
            tree, x[mask, :], k)  # 获得了各样本第k近邻的距离

    # 所有样本中的K近邻计算.
    tree = build_tree(x, metric)
    m = tree.query_radius(x, nn_distc_classes)
    m = [p.shape[0] for p in m]
    return (psi(N) - np.mean(psi(Nx_class)) + psi(k) - np.mean(psi(m))) / log(base)